@@ -1,4 +1,5 @@
 import os
+import numpy as np
 import re
 import random
 import pickle
@@ -13,7 +14,7 @@ try:
 except Exception:
     ImageSequenceClip = None
 
-from mohou.types import ImageBase, AngleVector, ElementDict, get_all_concrete_leaftypes
+from mohou.types import ImageBase, AngleVector, TerminateFlag, ElementDict, GripperState
 from mohou.types import MultiEpisodeChunk
 from mohou.model import AutoEncoder
 from mohou.model import AutoEncoderConfig
@@ -53,21 +54,17 @@ def create_default_logger(project_name: str, prefix: str) -> Logger:
 def train_autoencoder(
         project_name: str,
         image_type: Type[ImageBase],
-        use_aux_data: bool,
         model_config: AutoEncoderConfig,
         dataset_config: AutoEncoderDatasetConfig,
         train_config: TrainConfig,
         ae_type: Type[AutoEncoderBase] = AutoEncoder,
+        chunk: Optional[MultiEpisodeChunk] = None,
         warm_start: bool = False):
 
     logger = create_default_logger(project_name, 'autoencoder')
 
-    chunk = MultiEpisodeChunk.load(project_name)
-
-    if use_aux_data:
-        chunk_aux = MultiEpisodeChunk.load_aux(project_name)
-        chunk.merge(chunk_aux)
-        logger.info('aux data found and merged')
+    if chunk is None:
+        chunk = MultiEpisodeChunk.load(project_name)
 
     dataset = AutoEncoderDataset.from_chunk(chunk, image_type, dataset_config)
     if warm_start:
@@ -86,11 +83,14 @@ def train_lstm(
         model_config: LSTMConfig,
         dataset_config: AutoRegressiveDatasetConfig,
         train_config: TrainConfig,
+        chunk: Optional[MultiEpisodeChunk] = None,
         warm_start: bool = False):
 
     logger = create_default_logger(project_name, 'lstm')  # noqa
 
-    chunk = MultiEpisodeChunk.load(project_name)
+    if chunk is None:
+        chunk = MultiEpisodeChunk.load(project_name)
+
     dataset = AutoRegressiveDataset.from_chunk(chunk, embedding_rule, dataset_config)
     if warm_start:
         logger.info('warm start')
@@ -163,7 +163,8 @@ def add_text_to_image(image: ImageBase, text: str, color: str):
     ax = plt.subplot(1, 1, 1)
     ax.axis('off')
     ax.imshow(image.to_rgb()._data)
-    ax.text(7, 1, text, fontsize=25, color=color, verticalalignment='top')
+    bbox = dict(boxstyle='round', facecolor='white', alpha=0.7)
+    ax.text(7, 1, text, fontsize=15, color=color, verticalalignment='top', bbox=bbox)
     fig.canvas.draw()
     fig.canvas.flush_events()
     return canvas_to_ndarray(fig)
@@ -172,54 +173,99 @@ def add_text_to_image(image: ImageBase, text: str, color: str):
 def visualize_lstm_propagation(project_name: str, propagator: Propagator, n_prop: int):
 
     chunk = MultiEpisodeChunk.load(project_name).get_intact_chunk()
-    episode_data = chunk[0]
-
-    image_type = None
-    for key, embedder in propagator.embed_rule.items():
-        if issubclass(key, ImageBase):
-            image_type = key
-    assert image_type is not None
-
-    n_feed = 10
-    fed_avs = episode_data.get_sequence_by_type(AngleVector)[:n_feed]
-    fed_images = episode_data.get_sequence_by_type(image_type)[:n_feed]
-
-    print("start lstm propagation")
-    for elem_tuple in zip(fed_avs, fed_images):
-        propagator.feed(ElementDict(elem_tuple))
-    print("finish lstm propagation")
-
-    elem_dict_list = propagator.predict(n_prop)
-    pred_images = [elem_dict[image_type] for elem_dict in elem_dict_list]
-
-    print("adding text to images...")
-    fed_images_with_text = [add_text_to_image(image, 'fed (original)', 'blue') for image in fed_images]
-    pred_images_with_text = [add_text_to_image(image, 'predicted by lstm', 'green') for image in pred_images]
-
-    images_with_text = fed_images_with_text + pred_images_with_text
-
     save_dir = get_subproject_dir(project_name, 'lstm_result')
-    full_file_name = os.path.join(save_dir, 'result.gif')
-    assert ImageSequenceClip is not None, 'check if your moviepy is properly installed'
-    clip = ImageSequenceClip(images_with_text, fps=20)
-    clip.write_gif(full_file_name, fps=20)
 
+    for idx, edata in enumerate(chunk):
+        episode_data = chunk[idx]
 
-def auto_detect_autoencoder_type(project_name: str) -> Type[AutoEncoderBase]:
-    # TODO(HiroIshida) dirty...
-    t: Optional[Type[AutoEncoderBase]] = None
+        image_type = None
+        for key, embedder in propagator.embed_rule.items():
+            if issubclass(key, ImageBase):
+                image_type = key
+        assert image_type is not None
 
-    t_cand_list = get_all_concrete_leaftypes(AutoEncoderBase)
+        n_feed = 20
+        fed_avs = episode_data.get_sequence_by_type(AngleVector)[:n_feed]
+        fed_images = episode_data.get_sequence_by_type(image_type)[:n_feed]
 
-    detect_count = 0
-    for t_cand in t_cand_list:
-        try:
-            TrainCache.load(project_name, t_cand)
-            t = t_cand
-            detect_count += 1
-        except Exception:
-            pass
+        use_gripper = GripperState in propagator.embed_rule
 
-    assert detect_count == 1
-    assert t is not None  # redundant but for mypy check
-    return t
+        if use_gripper:
+            fed_grippers = episode_data.get_sequence_by_type(GripperState)[:n_feed]
+
+        print("start lstm propagation")
+        for i in range(n_feed):
+            elem_dict = ElementDict([fed_avs[i], fed_images[i]])
+            if use_gripper:
+                elem_dict[GripperState] = fed_grippers[i]
+            propagator.feed(elem_dict)
+        print("finish lstm propagation")
+
+        elem_dict_list = propagator.predict(n_prop)
+        pred_images = [elem_dict[image_type] for elem_dict in elem_dict_list]
+        pred_flags = [elem_dict[TerminateFlag].numpy().item() for elem_dict in elem_dict_list]
+        pred_avs = [elem_dict[AngleVector].numpy() for elem_dict in elem_dict_list]
+        if use_gripper:
+            pred_gss = [elem_dict[GripperState].numpy() for elem_dict in elem_dict_list]
+
+        n_av_dim = chunk.get_spec().type_shape_table[AngleVector][0]
+        n_gs_dim = chunk.get_spec().type_shape_table[GripperState][0] if use_gripper else 0
+        fig, axs = plt.subplots(n_av_dim + n_gs_dim, 1)
+
+        # plot angle vectors
+        av_seq_gt = episode_data.get_sequence_by_type(AngleVector)
+        np_av_seq_gt = np.array([av.numpy() for av in av_seq_gt])
+        np_av_seq_pred = np.concatenate((np_av_seq_gt[:n_feed], np.array(pred_avs)), axis=0)
+
+        i_dim = 0
+        for i_av_dim in range(n_av_dim):
+            axs[i_dim].plot(np_av_seq_gt[:, i_av_dim], color='blue', lw=1)
+            axs[i_dim].plot(np_av_seq_pred[:, i_av_dim], color='red', lw=1)
+
+            # determine axes min max
+            conc = np.hstack((np_av_seq_gt[:, i_av_dim], np_av_seq_pred[:, i_av_dim]))
+            y_min = np.min(conc)
+            y_max = np.max(conc)
+            diff = y_max - y_min
+            axs[i_dim].set_ylim([y_min - diff * 0.1, y_max + diff * 0.1])
+            axs[i_dim].set_title('AngleVector dim {}'.format(i_av_dim), fontsize=5, pad=0.0)
+            i_dim += 1
+
+        if use_gripper:
+            gs_seq_gt = episode_data.get_sequence_by_type(GripperState)
+            np_gs_seq_gt = np.array([gs.numpy() for gs in gs_seq_gt])
+            np_gs_seq_pred = np.concatenate((np_gs_seq_gt[:n_feed], np.array(pred_gss)), axis=0)
+            for i_gs_dim in range(n_gs_dim):
+                axs[i_dim].plot(np_gs_seq_gt[:, i_gs_dim], color='blue', lw=1)
+                axs[i_dim].plot(np_gs_seq_pred[:, i_gs_dim], color='red', lw=1)
+
+                # determine axes min max
+                conc = np.hstack((np_gs_seq_gt[:, i_gs_dim], np_gs_seq_pred[:, i_gs_dim]))
+                y_min = np.min(conc)
+                y_max = np.max(conc)
+                diff = y_max - y_min
+                axs[i_dim].set_ylim([y_min - diff * 0.1, y_max + diff * 0.1])
+                axs[i_dim].set_title('GripperState dim {}'.format(i_gs_dim), fontsize=5, pad=0.0)
+                i_dim += 1
+
+        for ax in axs:
+            ax.grid()
+
+        filename = os.path.join(save_dir, 'seq-{}{}.png'.format(AngleVector.__name__, idx))
+        fig.savefig(filename, format='png', dpi=300)
+        print('saved to {}'.format(filename))
+
+        # save gif image
+        print("adding text to images...")
+        fed_images_with_text = [add_text_to_image(image, 'fed (original) image)', 'blue') for image in fed_images]
+        clamp = lambda x: max(min(x, 1.0), 0.0)  # noqa
+        pred_images_with_text = [
+            add_text_to_image(image, 'predicted image (prob-terminated={:.2f})'.format(clamp(flag)), 'green')
+            for image, flag in zip(pred_images, pred_flags)]
+
+        images_with_text = fed_images_with_text + pred_images_with_text
+
+        full_file_name = os.path.join(save_dir, 'result-image{}.gif'.format(idx))
+        assert ImageSequenceClip is not None, 'check if your moviepy is properly installed'
+        clip = ImageSequenceClip(images_with_text, fps=20)
+        clip.write_gif(full_file_name, fps=20)
